import { Schema, ZodType, z } from 'zod'

import yaml from 'js-yaml'

import { readFile, writeFile } from 'fs/promises'
import { Mutex } from 'async-mutex'
import { debug } from './logger'

export const GitRef = z.union([
    z.object({
        type: z.literal('default'),
    }),
    z.object({
        type: z.union([z.literal('branch'), z.literal('branch'), z.literal('branch')]),
        value: z.string(),
    }),
])

export const GitDeploy = z.object({
    name: z.string(),
    url: z.string(),
    ref: GitRef,
})

const DockerRunOptions = z.object({
    name: z.string(),
    volumes: z.tuple([z.string(), z.string()]).array(),
    ports: z.tuple([z.string(), z.string()]).array(),
    env: z.record(z.string()),
})

export const DockerDeploy = z.object({
    name: z.string(),
    type: z.literal('docker'),
    options: DockerRunOptions.extend({
        image: z.string(),
    }),
})

export const DockerfileDeploy = GitDeploy.extend({
    type: z.literal('dockerfile'),
    options: DockerRunOptions.extend({
        path: z.string().optional(),
    }),
})

export const DockerComposeDeploy = GitDeploy.extend({
    type: z.literal('docker-compose'),
    options: z.object({
        path: z.string().optional(),
    }),
})

export const ShellDeploy = GitDeploy.extend({
    type: z.literal('shell'),
    options: z.object({
        path: z.string().optional(),
        env: z.record(z.string()),
    }),
})

export const Deploy = z.union([DockerDeploy, DockerfileDeploy, DockerComposeDeploy, ShellDeploy])

export type GitRef = z.infer<typeof GitRef>
export type GitDeploy = z.infer<typeof GitDeploy>

export type DockerDeploy = z.infer<typeof DockerDeploy>
export type DockerfileDeploy = z.infer<typeof DockerfileDeploy>
export type DockerComposeDeploy = z.infer<typeof DockerComposeDeploy>
export type ShellDeploy = z.infer<typeof ShellDeploy>

export type Deploy = z.infer<typeof Deploy>

export type Config = { deploys: Deploy[] }

export const refToString = (ref: GitRef, defaultStr: string = '<default>') => {
    return ref.type === 'default' ? defaultStr : ref.value
}

const mutex = new Mutex()

export function loadConfig(): Promise<Config> {
    return mutex.runExclusive(async () => {
        const data = await readFile('./config.yaml', 'utf8')
        return yaml.load(data) as Config
    })
}

export function updateConfig(fn: (config: Config) => Promise<void>) {
    return mutex.runExclusive(async () => {
        const data = await readFile('./config.yaml', 'utf8')
        const config = yaml.load(data) as Config

        await fn(config)

        await writeFile('./config.yaml', yaml.dump(config))
    })
}
