import child_process from 'child_process'

import { debug } from './logger'

type RunnerLogOptions = {
    silent?: boolean
}

export type Runner = {
    script: (source: string, options?: RunnerLogOptions) => Promise<void>
    command: (source: string, options?: RunnerLogOptions) => Promise<void>
}

function onLine(stream: any, cb: (line: string) => void) {
    let buffer = ''

    stream.on('data', (data: any) => {
        buffer += data
        let lines = buffer.split('\n')
        buffer = lines.pop()!
        lines.forEach(cb)
    })

    stream.on('end', () => {
        if (buffer) cb(buffer)
    })
}

export async function runShell(
    source: string,
    logLn: ((s: string) => void) | undefined
): Promise<void> {
    const child = child_process.exec(source)

    onLine(child.stdout!, line => {
        debug('[Runner]', ' >', line)
        logLn?.(line)
    })

    onLine(child.stderr!, line => {
        debug('[Runner]', '!>', line)
        logLn?.(line)
    })

    return new Promise((resolve, reject) => {
        child.on('close', code => {
            if (code === 0) {
                resolve()
            } else {
                reject(new Error(`Command failed with code ${code}`))
            }
        })
    })
}

export function createSimpleRunner(logLn: (s: string) => void): Runner {
    return {
        async command(source: string, { silent } = {}) {
            debug('[Runner] Command:')
            debug('[Runner]', ' |', source)

            await runShell(source, silent ? undefined : logLn)
        },
        async script(source: string, { silent } = {}) {
            debug('[Runner] Script:')
            source.split('\n').forEach(line => {
                debug('[Runner]', ' |', line)
            })

            await runShell(source, silent ? undefined : logLn)
        },
    }
}
